# Generated by Django 2.2.3 on 2019-08-15 13:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0006_remove_whitelist_and_alter_collection_version_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionversion',
            name='is_certified',
            field=models.BooleanField(default=False),
        ),
    ]
